@@ -57,6 +57,7 @@ is" without express or implied warranty.
 #include "Events.h"
 #include "Options.h"
 #include "Error.h"
+#include "Init.h"
 
 #include "compext/Compext.h"
 
@@ -78,10 +79,12 @@ is" without express or implied warranty.
 
 #include <errno.h>
 
-static int nxagentXkbGetNames(char **rules, char **model, char **layout,
-                                  char **variant, char **options);
+static void nxagentXkbGetNames(void);
 
-static void nxagentKeycodeConversionSetup(void);
+void nxagentKeycodeConversionSetup(void);
+
+static void nxagentWriteKeyboardDir(void);
+static void nxagentWriteKeyboardFile(char *rules, char *model, char *layout, char *variant, char *options);
 
 #endif /* XKB */
 
@@ -138,6 +141,13 @@ extern        Status        XkbGetControls(
 extern int XkbDfltRepeatDelay;
 extern int XkbDfltRepeatInterval;
 
+/* xkb configuration of the real X server */
+static char *nxagentRemoteRules = NULL;
+static char *nxagentRemoteModel = NULL;
+static char *nxagentRemoteLayout = NULL;
+static char *nxagentRemoteVariant = NULL;
+static char *nxagentRemoteOptions = NULL;
+
 #endif /* XKB */
 
 /*
@@ -569,8 +579,10 @@ int nxagentKeyboardProc(DeviceIntPtr pDev, int onoff)
   CARD8 modmap[MAP_LENGTH];
   int i, j;
   XKeyboardState values;
+#ifdef XKB
   char *model = NULL, *layout = NULL;
   XkbDescPtr xkb = NULL;
+#endif
 
   switch (onoff)
   {
@@ -692,25 +704,12 @@ N/A
       keySyms.mapWidth = mapWidth;
       keySyms.map = keymap;
 
-      if (XkbQueryExtension(nxagentDisplay,
-                            &nxagentXkbInfo.Opcode,
-                            &nxagentXkbInfo.EventBase,
-                            &nxagentXkbInfo.ErrorBase,
-                            &nxagentXkbInfo.MajorVersion,
-                            &nxagentXkbInfo.MinorVersion) == 0)
-      {
-          ErrorF("Unable to initialize XKEYBOARD extension.\n");
-          goto XkbError;
-      }
-
-
 #ifdef XKB
-
-      /*
-       * First of all the validity
-       * of XkbBaseDirectory global
-       * variable is checked.
-       */
+      if (!nxagentGetRemoteXkbExtension())
+      {
+        ErrorF("Unable to query XKEYBOARD extension.\n");
+        goto XkbError;
+      }
 
       if (noXkbExtension) {
         #ifdef TEST
@@ -744,7 +743,7 @@ XkbError:
 #ifdef XKB
       } else { /* if (noXkbExtension) */
         XkbComponentNamesRec names = {0};
-        char *rules = NULL, *variants = NULL, *options = NULL; /* use xkb default */
+        char *rules = NULL, *variant = NULL, *options = NULL; /* use xkb default */
 
         #ifdef TEST
         fprintf(stderr, "nxagentKeyboardProc: Using XKB extension.\n");
@@ -754,6 +753,15 @@ XkbError:
         fprintf(stderr, "nxagentKeyboardProc: nxagentKeyboard is [%s].\n", nxagentKeyboard ? nxagentKeyboard : "NULL");
         #endif
 
+        if (nxagentX2go == 1 && nxagentKeyboard && (strcmp(nxagentKeyboard, "null/null") == 0))
+        {
+          #ifdef TEST
+          fprintf(stderr, "%s: changing nxagentKeyboard from [null/null] to [clone].\n", __func__);
+          #endif
+          free(nxagentKeyboard);
+          nxagentKeyboard = strdup("clone");
+        }
+
         /*
           from nxagent changelog:
           2.0.22:
@@ -765,7 +773,9 @@ XkbError:
           layout.
         */
 
-        if (nxagentKeyboard && (strcmp(nxagentKeyboard, "query") != 0))
+        if (nxagentKeyboard &&
+            (strcmp(nxagentKeyboard, "query") != 0) &&
+            (strcmp(nxagentKeyboard, "clone") != 0))
         {
           for (i = 0; nxagentKeyboard[i] != '/' && nxagentKeyboard[i] != 0; i++);
 
@@ -776,8 +786,33 @@ XkbError:
             goto XkbError;
           }
 
-          model = strndup(nxagentKeyboard, i);
-          layout = strdup(&nxagentKeyboard[i + 1]);
+          /*
+            The original nxagent only supports model/layout values
+            here. It uses these values together with the default rules
+            and empty variant and options. We use a more or less
+            compatible hack here: The special keyword rlmvo for model
+            means that the layout part of the string will contain a
+            full RMLVO config, separated by #, e.g.
+            rlmvo/base#pc105#de,us#nodeadkeys#lv3:rwin_switch
+          */
+          if (strncmp(nxagentKeyboard, "rlmvo/", 6) == 0)
+          {
+            const char * sep = "#";
+            char * rmlvo = strdup(&nxagentKeyboard[i+1]);
+            char * tmp = rmlvo;
+            /* strtok cannot handle empty fields, so use strsep */
+            rules = strdup(strsep(&tmp, sep));
+            model = strdup(strsep(&tmp, sep));
+            layout = strdup(strsep(&tmp, sep));
+            variant = strdup(strsep(&tmp, sep));
+            options = strdup(strsep(&tmp, sep));
+            free(rmlvo);
+          }
+          else
+          {
+            model = strndup(nxagentKeyboard, i);
+            layout = strdup(&nxagentKeyboard[i + 1]);
+          }
 
           /*
            * There is no description for pc105 on Solaris.
@@ -785,8 +820,8 @@ XkbError:
            */
 
           #ifdef TEST
-          fprintf(stderr, "nxagentKeyboardProc: Using keyboard model [%s] with layout [%s].\n",
-                      model, layout);
+          fprintf(stderr, "%s: Using [rules='%s',model='%s',layout='%s',variant='%s',options='%s'].\n",
+                  __func__, rules, model, layout, variant, options);
           #endif
 
           #ifdef __sun
@@ -795,7 +830,6 @@ XkbError:
           {
             #ifdef TEST
             fprintf(stderr, "nxagentKeyboardProc: WARNING! Keyboard model 'pc105' unsupported on Solaris.\n");
-
             fprintf(stderr, "nxagentKeyboardProc: WARNING! Forcing keyboard model to 'pc104'.\n");
             #endif
 
@@ -816,9 +850,59 @@ XkbError:
         fprintf(stderr, "nxagentKeyboardProc: Init XKB extension.\n");
         #endif
 
-        xkb = XkbGetKeyboard(nxagentDisplay, XkbGBN_AllComponentsMask, XkbUseCoreKbd);
+        if (nxagentRemoteRules && nxagentRemoteModel)
+        {
+          #ifdef DEBUG
+          fprintf(stderr, "%s: Remote: [rules='%s',model='%s',layout='%s',variant='%s',options='%s'].\n",
+                  __func__, nxagentRemoteRules, nxagentRemoteModel, nxagentRemoteLayout, nxagentRemoteVariant, nxagentRemoteOptions);
+          #endif
 
-        nxagentKeycodeConversionSetup();
+
+          /* Only setup keycode conversion if we are NOT in clone mode */
+          if (nxagentKeyboard && (strcmp(nxagentKeyboard, "clone") == 0))
+          {
+            free(rules); rules = strdup(nxagentRemoteRules);
+            free(model); model = strdup(nxagentRemoteModel);
+            free(layout); layout = strdup(nxagentRemoteLayout);
+            free(variant); variant = strdup(nxagentRemoteVariant);
+            free(options); options = strdup(nxagentRemoteOptions);
+            /*
+             * when cloning we do not want X2Go to set the keyboard
+             * via a keyboard file generated by nxagent. The defined
+             * method for switching that off is the creation of a dir
+             * instead of a file. Which is achieved by passing NULL to
+             * nxagentWriteKeyboardFile.
+             */
+            if (nxagentX2go == 1)
+              nxagentWriteKeyboardDir();
+          }
+          else
+          {
+            nxagentKeycodeConversionSetup();
+            /*
+             * Keyboard has always been tricky with nxagent. For that
+             * reason X2Go offers "auto" keyboard configuration. You can
+             * specify it in the client side session configuration. In
+             * "auto" mode x2goserver expects nxagent to write the
+             * remote keyboard config to a file on startup and
+             * x2goserver would then pick that file and pass it to
+             * setxkbmap. This functionality is obsoleted by the "clone"
+             * stuff but we still need it because x2goserver does not
+             * know about that yet. Once x2go starts using clone
+             * we can drop this here.
+             */
+            if (nxagentX2go == 1)
+              nxagentWriteKeyboardFile(nxagentRemoteRules, nxagentRemoteModel, nxagentRemoteLayout, nxagentRemoteVariant, nxagentRemoteOptions);
+          }
+        }
+        #ifdef DEBUG
+        else
+        {
+          fprintf(stderr, "%s: Failed to retrieve remote rules.\n", __func__);
+        }
+        #endif
+
+        xkb = XkbGetKeyboard(nxagentDisplay, XkbGBN_AllComponentsMask, XkbUseCoreKbd);
 
         if (xkb && xkb->geom)
         {
@@ -833,11 +917,11 @@ XkbError:
 
         #ifdef DEBUG
         fprintf(stderr, "nxagentKeyboardProc: Going to set rules and init device: "
-                        "[rules='%s',model='%s',layout='%s',variants='%s',options='%s'].\n",
-                        rules, model, layout, variants, options);
+                        "[rules='%s',model='%s',layout='%s',variant='%s',options='%s'].\n",
+                        rules, model, layout, variant, options);
         #endif
 
-        XkbSetRulesDflts(rules, model, layout, variants, options);
+        XkbSetRulesDflts(rules, model, layout, variant, options);
         XkbInitKeyboardDeviceStruct((void *)pDev, &names, &keySyms, modmap,
                                     nxagentBell, nxagentChangeKeyboardControl);
 
@@ -1473,8 +1557,16 @@ void nxagentTuneXkbWrapper(void)
   }
 }
 
-static int nxagentXkbGetNames(char **rules, char **model, char **layout,
-                                  char **variant, char **options)
+void nxagentXkbClearNames(void)
+{
+  free(nxagentRemoteRules); nxagentRemoteRules = NULL;
+  free(nxagentRemoteModel); nxagentRemoteModel = NULL;
+  free(nxagentRemoteLayout); nxagentRemoteLayout = NULL;
+  free(nxagentRemoteVariant); nxagentRemoteVariant = NULL;
+  free(nxagentRemoteOptions); nxagentRemoteOptions = NULL;
+}
+
+static void nxagentXkbGetNames(void)
 {
   Atom atom;
   #ifdef _XSERVER64
@@ -1489,28 +1581,25 @@ static int nxagentXkbGetNames(char **rules, char **model, char **layout,
   char *name;
   Status result;
 
-  data = name = NULL;
-
-  *rules = NULL;
-  *model = NULL;
-  *layout = NULL;
-  *variant = NULL;
-  *options = NULL;
+  if (nxagentRemoteRules)
+    return;
 
   atom = XInternAtom(nxagentDisplay, "_XKB_RULES_NAMES", 1);
 
   if (atom == 0)
   {
-    return 0; 
+    return;
   }
 
+  data = name = NULL;
+
   result = XGetWindowProperty(nxagentDisplay, DefaultRootWindow(nxagentDisplay),
                                   atom, 0, 256, 0, XA_STRING, &type, &format,
                                       &n, &after, (unsigned char **)&data);
 
   if (result != Success || !data)
   {
-    return 0;
+    return;
   }
 
   if ((after > 0) || (type != XA_STRING) || (format != 8))
@@ -1518,7 +1607,7 @@ static int nxagentXkbGetNames(char **rules, char **model, char **layout,
     if (data)
     {
       XFree(data);
-      return 0;
+      return;
     }
   }
 
@@ -1526,103 +1615,116 @@ static int nxagentXkbGetNames(char **rules, char **model, char **layout,
 
   if (name < data + n)
   {
-    *rules = name;
+    nxagentRemoteRules = strdup(name);
     name += strlen(name) + 1;
   }
 
   if (name < data + n)
   {
-    *model = name;
+    nxagentRemoteModel = strdup(name);
     name += strlen(name) + 1;
   }
 
   if (name < data + n)
   {
-    *layout = name;
+    nxagentRemoteLayout = strdup(name);
     name += strlen(name) + 1;
   }
 
   if (name < data + n)
   {
-    *variant = name;
+    nxagentRemoteVariant = strdup(name);
     name += strlen(name) + 1;
   }
 
   if (name < data + n)
   {
-    *options = name;
+    nxagentRemoteOptions = strdup(name);
     name += strlen(name) + 1;
   }
 
-  return n;
+  XFree(data);
+
+  return;
 }
 
-void nxagentKeycodeConversionSetup(void)
+static void writeKeyboardfileData(FILE *out, char *rules, char *model, char *layout, char *variant, char *options)
 {
-  char *drules = NULL;
-  char *dmodel = NULL;
-  char *dlayout = NULL;
-  char *dvariant = NULL;
-  char *doptions = NULL;
-  unsigned int drulesLen;
+  /*
+    How to set "empty" values with setxkbmap, result of trial and error:
+    - model and layout: empty strings are accepted by setxkbmap.
+    - rules: setxkbmap will fail if rules is an empty string
+      (code will intercept in an earlier stage in that case)
+    - variant: the variant line must be omitted completely.
+    - options: prepend value with "," to override, otherwise options will be added.
+  */
+  fprintf(out, "rules=\"%s\"\n", rules);
+  fprintf(out, "model=\"%s\"\n", model ? model : "");
+  fprintf(out, "layout=\"%s\"\n", layout ? layout : "");
+  if (variant && variant[0] != '\0')
+    fprintf(out, "variant=\"%s\"\n", variant);
+  fprintf(out, "options=\",%s\"\n", options ? options : "");
+}
 
-  if (nxagentOption(KeycodeConversion) == KeycodeConversionOff)
+static char* getKeyboardFilePath(void)
+{
+  char *keyboard_file_path = NULL;
+  char *sessionpath = nxagentGetSessionPath();
+  if (sessionpath)
   {
-    fprintf(stderr, "Info: Keycode conversion is off\n");
-    nxagentKeycodeConversion = False;
-    return;
-  }
-  else if (nxagentOption(KeycodeConversion) == KeycodeConversionOn)
-  {
-    fprintf(stderr, "Info: Keycode conversion is on\n");
-    nxagentKeycodeConversion = True;
-    return;
-  }
-
-  nxagentKeycodeConversion = False;
-
-  drulesLen = nxagentXkbGetNames(&drules, &dmodel, &dlayout,
-                                     &dvariant, &doptions);
-
-  #ifdef DEBUG
-  if (drulesLen != 0 && drules && dmodel)
-  {
-    fprintf(stderr, "nxagentKeycodeConversionSetup: "
-                    "Remote: [rules='%s',model='%s',layout='%s',variant='%s',options='%s'].\n",
-                    drules, dmodel, dlayout, dvariant, doptions);
+    if ((asprintf(&keyboard_file_path, "%s/keyboard", sessionpath) == -1))
+    {
+      free(sessionpath);
+      FatalError("malloc for keyboard file path failed.");
+    }
+    free(sessionpath);
   }
   else
   {
-    fprintf(stderr, "nxagentKeycodeConversionSetup: "
-                "Failed to retrieve remote rules.\n");
+    fprintf(stderr, "Warning: Failed to determine keyboard file path: SessionPath not defined\n");
   }
-  #endif
+  return keyboard_file_path;
+}
 
-  if (drulesLen != 0)
+static void nxagentWriteKeyboardDir(void)
+{
+  char *keyboard_file_path = getKeyboardFilePath();
+
+  if (keyboard_file_path)
   {
-    char *sessionpath = nxagentGetSessionPath();
-    if (sessionpath)
+    /*
+     * special case: if rules is NULL create a directory insteas of
+     * a file. This is the defined method to disable x2gosetkeyboard.
+     */
+    if (mkdir(keyboard_file_path, 0555) < 0)
+    {
+      int save_err = errno;
+      fprintf(stderr, "Warning: Failed to create keyboard blocking directory '%s': %s\n", keyboard_file_path, strerror(save_err));
+    }
+    else
+    {
+      fprintf(stderr, "Info: keyboard blocking directory created: '%s'\n", keyboard_file_path);
+    }
+    free(keyboard_file_path);
+  }
+}
+
+static void nxagentWriteKeyboardFile(char *rules, char *model, char *layout, char *variant, char *options)
+{
+  if (rules && rules[0] != '\0')
+  {
+    #ifdef DEBUG
+    writeKeyboardfileData(stderr, rules, model, layout, variant, options);
+    #endif
+
+    char *keyboard_file_path = getKeyboardFilePath();
+
+    if (keyboard_file_path)
     {
-      char *keyboard_file_path = NULL;
       FILE *keyboard_file;
-      if ((asprintf(&keyboard_file_path, "%s/keyboard", sessionpath) == -1))
-      {
-        free(sessionpath);
-        FatalError("malloc for keyboard file path failed.");
-      }
-      free(sessionpath);
       if ((keyboard_file = fopen(keyboard_file_path, "w")))
       {
-        if (drules)
-          fprintf(keyboard_file, "rules=\"%s\"\n", drules[0] == '\0' ? "," : drules);
-        if (dmodel)
-          fprintf(keyboard_file, "model=\"%s\"\n", dmodel[0] == '\0' ? "," : dmodel);
-        if (dlayout)
-          fprintf(keyboard_file, "layout=\"%s\"\n", dlayout[0] == '\0' ? "," : dlayout);
-        if (dvariant)
-          fprintf(keyboard_file, "variant=\"%s\"\n", dvariant[0] == '\0' ? "," : dvariant);
-        if (doptions)
-          fprintf(keyboard_file, "options=\"%s\"\n", doptions[0] == '\0' ? "," : doptions);
+        writeKeyboardfileData(keyboard_file, rules, model, layout, variant, options);
         fclose(keyboard_file);
         fprintf(stderr, "Info: keyboard file created: '%s'\n", keyboard_file_path);
       }
@@ -1633,61 +1735,72 @@ void nxagentKeycodeConversionSetup(void)
       }
       free(keyboard_file_path);
     }
-    else
-    {
-      fprintf(stderr, "Warning: SessionPath not defined\n");
-    }
   }
-  else
+}
+
+void nxagentKeycodeConversionSetup(void)
+{
+  nxagentKeycodeConversion = False;
+
+  if (nxagentXkbInfo.Opcode == -1)
+    return;
+
+  if (nxagentOption(KeycodeConversion) == KeycodeConversionOff)
   {
-    fprintf(stderr, "Warning: Failed to create the keyboard file\n");
+    fprintf(stderr, "Info: Keycode conversion is off\n");
   }
-
-  if (drules && dmodel &&
-      (strcmp(drules, "evdev") == 0 ||
-       strcmp(dmodel, "evdev") == 0))
+  else if (nxagentOption(KeycodeConversion) == KeycodeConversionOn)
   {
-    #ifdef DEBUG
-    fprintf(stderr, "nxagentKeycodeConversionSetup: "
-                "Activating KeyCode conversion.\n");
-    #endif
-
-    fprintf(stderr, "Info: Keycode conversion auto-determined as on\n");
+    fprintf(stderr, "Info: Keycode conversion is on\n");
     nxagentKeycodeConversion = True;
   }
   else
   {
-    fprintf(stderr, "Info: Keycode conversion auto-determined as off\n");
-  }
+    if (nxagentRemoteRules && nxagentRemoteModel &&
+        (strcmp(nxagentRemoteRules, "evdev") == 0 ||
+         strcmp(nxagentRemoteModel, "evdev") == 0))
+    {
+      #ifdef DEBUG
+      fprintf(stderr, "%s: Activating KeyCode conversion.\n", __func__);
+      #endif
 
-  if (drules)
-  {
-    XFree(drules);
+      fprintf(stderr, "Info: Keycode conversion auto-determined as on\n");
+      nxagentKeycodeConversion = True;
+    }
+    else
+    {
+      #ifdef DEBUG
+      fprintf(stderr, "%s: Deactivating KeyCode conversion.\n", __func__);
+      #endif
+
+      fprintf(stderr, "Info: Keycode conversion auto-determined as off\n");
+    }
   }
 }
 
-void nxagentResetKeycodeConversion(void)
+Bool nxagentGetRemoteXkbExtension(void)
 {
-  int result;
-  XkbAgentInfoRec info;
+  Bool result;
 
-  result = XkbQueryExtension(nxagentDisplay, &info.Opcode, &info.EventBase,
-                                 &info.ErrorBase, &info.MajorVersion,
-                                     &info.MinorVersion);
+  nxagentXkbInfo.Opcode = nxagentXkbInfo.EventBase = nxagentXkbInfo.ErrorBase = nxagentXkbInfo.MajorVersion = nxagentXkbInfo.MinorVersion = -1;
+  nxagentXkbClearNames();
 
-  if (result != 0)
+  if ((result = XkbQueryExtension(nxagentDisplay,
+                                 &nxagentXkbInfo.Opcode,
+                                 &nxagentXkbInfo.EventBase,
+                                 &nxagentXkbInfo.ErrorBase,
+                                 &nxagentXkbInfo.MajorVersion,
+                                 &nxagentXkbInfo.MinorVersion)))
   {
-    nxagentKeycodeConversionSetup();
+    nxagentXkbGetNames();
   }
+  #ifdef WARNING
   else
   {
-    #ifdef WARNING
-    fprintf(stderr, "nxagentResetKeycodeConversion: "
-                "WARNING! Failed to query XKB extension.\n");
-    #endif
-
-    nxagentKeycodeConversion = False;
+    fprintf(stderr, "%s: WARNING! Failed to query XKB extension.\n", __func__);
   }
-}
+  #endif
 
+  return result;
+}
 #endif /* XKB */
