@@ -1346,7 +1346,12 @@ ProcRenderCompositeGlyphs (ClientPtr client)
     {
 	listsBase = (GlyphListPtr) malloc (nlist * sizeof (GlyphListRec));
 	if (!listsBase)
+	{
+	    free(glyphsBase);
+	    free(listsBase);
+
 	    return BadAlloc;
+	}
     }
     buffer = (CARD8 *) (stuff + 1);
     glyphs = glyphsBase;
@@ -2918,9 +2923,9 @@ PanoramiXRenderFillRectangles (ClientPtr client)
 	    result = (*PanoramiXSaveRenderVector[X_RenderFillRectangles]) (client);
 	    if(result != Success) break;
 	}
+
 	free(extra);
     }
-
     return result;
 }
 
@@ -2979,8 +2984,8 @@ PanoramiXRenderTrapezoids(ClientPtr client)
 
 	    if(result != Success) break;
 	}
-	
-        free(extra);
+
+	free(extra);
     }
 
     return result;
@@ -3037,8 +3042,8 @@ PanoramiXRenderTriangles(ClientPtr client)
 
 	    if(result != Success) break;
 	}
-	
-        free(extra);
+
+	free(extra);
     }
 
     return result;
@@ -3091,8 +3096,8 @@ PanoramiXRenderTriStrip(ClientPtr client)
 
 	    if(result != Success) break;
 	}
-	
-        free(extra);
+
+	free(extra);
     }
 
     return result;
@@ -3145,8 +3150,8 @@ PanoramiXRenderTriFan(ClientPtr client)
 
 	    if(result != Success) break;
 	}
-	
-        free(extra);
+
+	free(extra);
     }
 
     return result;
@@ -3276,6 +3281,7 @@ PanoramiXRenderAddTraps (ClientPtr client)
 	    result = (*PanoramiXSaveRenderVector[X_RenderAddTraps]) (client);
 	    if(result != Success) break;
 	}
+
 	free(extra);
     }
 
